source("aux_functions.R")
bernoulli <- function(p){
  u <- runif(1)
  if (u<p){
    x=0
  }else{
    x=1
  } 
  return(x)
}

#temp <- get_simulation_huge(100,m=12)
BVAR <- function(Yraw,nsave,nburn,p,h,a_i=0.01,b_i=0.01,var.c1= .01,var.c2 = .01,prmean=1,cons=1,fcst=1,fhorz=4,imps=0,imphor=40,sv=TRUE,shrink.type="Minnesota",store.draws=FALSE,flat=FALSE,method="LIK"){
  #-------------------------------------------Load packages -----------------------------------------------------------------#
  require(GIGrvg)
  require(Matrix)
  require(MASS)
  require(mvtnorm)
  require(snowfall)
  require(stochvol)
  mlag <- function(X,lag){
    p <- lag
    X <- as.matrix(X)
    Traw <- nrow(X)
    N <- ncol(X)
    Xlag <- matrix(0,Traw,p*N)
    for (ii in 1:p){
      Xlag[(p+1):Traw,(N*(ii-1)+1):(N*ii)]=X[(p+1-ii):(Traw-ii),(1:N)]
    }
    return(Xlag)  
  }  
  #FUNCTION INPUT
  #Yraw <- new.data$small;nsave <- 5000;nburn <- 5000;p <- 5;h <- 0;a_tau <- 0.1;b_tau <- 0.1;c_tau=1.5;d_tau=1;a_i=0.01;b_i=0.01;var.c1= .01;var.c2 = .01;prmean=1;cons=1;fcst=1;fhorz=4;sample_a=1;imps=0;imphor=40;sv=TRUE;shrink.type="Minnesota";store.draws=TRUE;
  
  #-------------------------------------------Some settings-----------------------------------------------------------------#
  #Number of saved draws and burn-ins
  ntot <- nsave+nburn
  #-----------------------------------------Data preliminaries---------------------------------------------------------------#
  if (cons==1){
    Xraw <- cbind(mlag(Yraw,p),1)
  }else{
    Xraw <- cbind(mlag(Yraw,p))
  }
  Y <- Yraw[(p+1):(nrow(Yraw)-h),]
  X <- Xraw[(p+1):(nrow(Xraw)-h),]
  T <- nrow(X)
  M <<- ncol(Y)
  K <<- ncol(X)
  k <- K*M
  v <- (M*(M-1))/2
  #-----------------------------------------Initialize Gibbs sampler---------------------------------------------------------#
  if (shrink.type=="SSVS"){
    #---------------------------------------------------PRIOR Hyperparameters--------------------------------------
    n <- K*M
    gamma_draws <- matrix(0,nsave,n)
    #------------------------------------------------PRIOR SETUP FOR SSVS PRIOR-----------------------------------#
    # Priors on SIGMA
    # SSVS variances for non-diagonal elements of SIGMA     
    kappa_0  <-  .1 #Set kappa_[0ij], kappa_[1ij]
    kappa_1  <-  7
    # Hyperparameters for Gamma ~ BERNOULLI(m,p_i), see eq. (14)
    p_i  <-  .5
    
    # Hyperparameters for Omega_[j] ~ BERNOULLI(j,q_ij), see eq. (17)
    q_ij  <- .5
    
    # Initialize Gamma and Omega vectors
    gammas  <-  matrix(1,n,1)       # vector of Gamma
    omega <- list() #!!!!!!!
    for (kk_1 in 1:(M-1)){
      omega[[kk_1]]  <-  matrix(1,kk_1,1)  # Omega_j
    }
    #--------------------------------------------PRIOR Calibration ----------------------------------------------
    #------------------------------------------getting OLS estimates--------------------------------------------- 
    XtXinv <- ginv(crossprod(X))
    A_OLS <- A_draw <-  XtXinv%*%(t(X)%*%Y)
    a_OLS <- as.vector(A_OLS) #vectorizes A_OLS, i.e. a_OLS=vec(A_OLS)
    Em <- Em.str <- Y-X%*%A_OLS
    SSE  <-  t((Y - X%*%A_OLS))%*%(Y - X%*%A_OLS)
    SIGMA_OLS  <-  SSE/ifelse((T-K+1)<0, 1, T-K+1)
    #-------------Initialize Bayesian Posteriors using OLS values-
    alpha <- a_OLS # single draw from the posterior of alpha
    ALPHA <- A_OLS # -,,- ALPHA
    SSE_Gibbs <- SSE # of SSE
    SIGMA <- SIGMA_OLS #of SIGMA
    
    A_prior <- matrix(0,K,M)
    A_prior[1:M,1:M] <- diag(M)*prmean

#a_prior  <-  matrix(0,n,1)
    
    # This is the std of the OLS estimate of alpha. You can use this to 
    # scale tau_0 and tau_1 (see below) if you want.
    # SSVS variances for alpha
    sigma_alpha  <-  sqrt(diag(kronecker(SIGMA_OLS,XtXinv)))
    if (!is.finite(sum(sigma_alpha))) sigma_alpha <- matrix(1,n,1)
    
    
    tau_0  <-  .1*sigma_alpha   # Set tau_[0i], tau_[1i]
    tau_1  <-  10*sigma_alpha
  }else{
    #---------------------------------------------------PRIOR Hyperparameters--------------------------------------
    n <- K*M
    #--------------------------------------------PRIOR Calibration ----------------------------------------------
    #------------------------------------------getting OLS estimates--------------------------------------------- 
    XtXinv <- ginv(crossprod(X))
    A_OLS <- A_draw <-  XtXinv%*%(t(X)%*%Y)
    a_OLS <- as.vector(A_OLS) #vectorizes A_OLS, i.e. a_OLS=vec(A_OLS)
    Em <-Em.str <-  Y-X%*%A_OLS
    SSE  <-  t((Y - X%*%A_OLS))%*%(Y - X%*%A_OLS)
    SIGMA_OLS  <-  SSE/(T-K+1)
    #-------------Initialize Bayesian Posteriors using OLS values-
    alpha <- a_OLS # single draw from the posterior of alpha
    ALPHA <- A_OLS # -,,- ALPHA
    SSE_Gibbs <- SSE # of SSE
    SIGMA <- SIGMA_OLS #of SIGMA
    
    lambdas <- seq(1,4)
    ind <- matrix(0,M,p)
    for (i in 1:M){
      ind[i,] <- seq(i,K-cons, by=M)
    }
    #Now get residual variances of univariate p-lag AR models. 
    #AR(p) is run on each equation, ignoring the constant and exogenous variables
    #-------------------------------------------------------------------------
    sigma_sq  <- matrix(0,M,1) #vector which stores the residual variance
    
    for (i in 1:M){
      Ylag_i <- mlag(Yraw[,i],p)
      Ylag_i <- Ylag_i[(p+1):nrow(Ylag_i),]
      Y_i <- Yraw[(p+1):nrow(Yraw),i]
      alpha_i <- solve(t(Ylag_i)%*%Ylag_i)%*%t(Ylag_i)%*%Y_i
      sigma_sq[i,1] <- (1/(nrow(Yraw)-p))*t(Y_i-Ylag_i%*%alpha_i)%*%(Y_i-Ylag_i%*%alpha_i) 
    }

    #-----------------PRIOR HYPERPARAMETERS-------------------
    #creating an array of dimension K x N, which will contain the K diagonal elements of the covmatrix, in each eqn
    A_prior <- matrix(0,K,M)
    A_prior[1:M,1:M] <- diag(M)*prmean

    
    shrink.1 <- 0.1
    shrink.2 <- 0.2
    if (shrink.type=="RW"){
     shrink.1<- shrink.2 <-1/10^8
     A_prior <- matrix(0,K,M)
     A_prior[1:M,1:M] <- diag(M)*1

    }
    if (flat){
      shrink.1 <- 10^5
      shrink.2 <- 10^5
    }
    
    V_i <- get_V(shrink.1,shrink.2,ind1=ind,sigma_sq1=sigma_sq,p1=p,set.cons=cons,M1=M,K1=K) 
    rownames(V_i) <- colnames(X)
    
    V_prior <- diag(as.vector(V_i))
    
  }
  D <- diag(diag(SIGMA))
 
  a_prior <- as.vector(A_prior)
  
  beta_prior <- matrix(0,v,1)
  if (shrink.type=="SSVS") theta <- matrix(10,K,M) else theta <- V_i
  omega <- matrix(10^2,v,1)
  b_draw <- lower.tri(SIGMA_OLS)*10
  if (shrink.type=="RW") b_draw <- lower.tri(SIGMA_OLS)*1/10^8
  
  Upsilon_draw <- b_draw
  
  diag(Upsilon_draw) <- 1
  
  #SV
  svdraw <- list(para=c(mu=-10,phi=.9,sigma=.2),latent=rep(-3,T))
  svl <- list()
  for (jj in 1:M) svl[[jj]] <- svdraw
  
  hv <- svdraw$latent
  para <- list(mu=-3,phi=.9,sigma=.2)
  H <- matrix(-3,T,M)
  eta <- list()
  #MH preliminaries
  accept <- 0
  accept2 <- 0
  scale1 <- .43
  scale2 <- .45
  #------------------------------------------Create storage matrices----------------------------------------------------------#
  if (store.draws){
    A_store <- array(0,c(nsave,K,M))
    SIGMA_store <- array(0,c(nsave,M,M))
    theta_store <- array(0,c(nsave,K,M))
    omega_store <- array(0,c(nsave,M,M))
    Upsilon_store <- array(0,c(nsave,M,M))
    hv_store <- array(0,c(nsave,T,M))
    gamma_store <- array(0,c(nsave,K,M))
    shrink_store <- array(0,c(nsave,2))
    if (p==1) IRF_store <- array(NA,c(nsave,M,M,imphor-1)) else IRF_store <- array(NA,c(nsave,M,M,imphor))
  }
  pred_store <- array(0,c(nsave,M,fhorz))
  a_store <- matrix(0,nsave,2)
  Lik_store <- matrix(0,nsave,1)
  pars_store <- array(0,c(nsave,3,M))
  
  #--------------------------------------------Start big Gibbs loop-----------------------------------------------------------#
  # pb <- txtProgressBar(min = 0, max = ntot, style = 3) #start progress bar
  start <- Sys.time()
  for (irep in 1:ntot){
     #Step I: Sample autoregressive parameters
    for (mm in 1:M){
        #Sample autoregressive coefficients equation by equation
        if (mm==1){
          Y.i <- Y[,mm]*exp(-0.5*H[,mm])
          X.i <- X*exp(-0.5*H[,mm])
          
          V_post <- try(chol2inv(chol(crossprod(X.i)+diag(1/theta[,mm]))),silent=TRUE)
          if (is(V_post,"try-error")) V_post <- ginv(crossprod(X.i)+diag(1/theta[,mm]))
          A_post <- V_post%*%(crossprod(X.i,Y.i)+diag(1/theta[,mm])%*%A_prior[,mm])
          
          A.draw.i <- try(A_post+t(chol(V_post))%*%rnorm(ncol(X.i)),silent=TRUE)
          if (is(A.draw.i,"try-error")) A.draw.i <- mvrnorm(1,A_post,V_post)
          A_draw[,mm] <- A.draw.i
          Em[,mm] <- Em.str[,mm] <- Y[,mm]-X%*%A.draw.i
        }else{
          Y.i <- Y[,mm]*exp(-0.5*H[,mm])
          X.i <- cbind(X,Em[,1:(mm-1)])*exp(-0.5*H[,mm])
          
          V_post <- try(chol2inv(chol((crossprod(X.i)+diag(1/c(theta[,mm],b_draw[mm,1:(mm-1)]))))),silent=TRUE)    #solve(crossprod(X.i)+diag(1/c(theta[,mm],b_draw[mm,1:(mm-1)])))
          if (is(V_post,"try-error")) V_post <- ginv((crossprod(X.i)+diag(1/c(theta[,mm],b_draw[mm,1:(mm-1)]))))
          A_post <- V_post%*%(crossprod(X.i,Y.i)+diag(1/c(theta[,mm],b_draw[mm,1:(mm-1)]))%*%c(A_prior[,mm],rep(0,mm-1)))
          
          A.draw.i <- try(A_post+t(chol(V_post))%*%rnorm(ncol(X.i)),silent=TRUE)
          if (is(A.draw.i,"try-error")) A.draw.i <- mvrnorm(1,A_post,V_post)
          
          A_draw[,mm] <- A.draw.i[1:ncol(X)]
          Em[,mm] <- Y[,mm]-X%*%A.draw.i[1:ncol(X)]
          Em.str[,mm] <- Y[,mm]-X%*%A.draw.i[1:ncol(X)]-Em[,1:(mm-1),drop=FALSE]%*%A.draw.i[(ncol(X)+1):ncol(X.i),drop=FALSE] 

          Upsilon_draw[mm,1:(mm-1)] <- eta[[mm]]<- A.draw.i[(ncol(X)+1):ncol(X.i)] 
        }
    }
    
  if (shrink.type=="SSVS"){
    alpha <- as.vector(A_draw)
    # Draw gamma|phi,psi,eta,omega,DATA from BERNOULLI dist.    
    for (nn_6 in 1:n){
      u_i1  <-  (1/tau_0[nn_6])*exp(-0.5*(alpha[nn_6]/(tau_0[nn_6]))^2)*p_i        
      u_i2  <-  (1/tau_1[nn_6])*exp(-0.5*(alpha[nn_6]/(tau_1[nn_6]))^2)*(1-p_i)
      gst  <-  u_i1/(u_i1 + u_i2)
      if (gst=="NaN") gst <- 0 
      gammas[nn_6,1]  <-  bernoulli(gst) 
      gammas[is.na(gammas)] <- 1
    }
    theta <-  matrix(0,n,1)   # h_i is tau_0 if gamma=0 and tau_1 if gamma=1
    for (nn_3 in 1:n){
      if (gammas[nn_3,1] == 0){               
        theta[nn_3,1]  <-  tau_0[nn_3]^2
      }else if (gammas[nn_3,1] == 1){        
        theta[nn_3,1]  <-  tau_1[nn_3]^2     
      }
     }
    
    theta <- matrix(theta,K,M)
    
  }else if (shrink.type=="Minnesota"){
    if (!flat){
    #Step for the first shrinkage parameter
    shrink.1.prop <- exp(rnorm(1,0,scale1))*shrink.1
    theta.prop <- get_V(shrink.1.prop,shrink.2,ind1=ind,sigma_sq1=sigma_sq,p1=p,set.cons=cons,M1=M,K1=K)
    
    post1.prop <- sum(dnorm(as.vector(A_draw),as.vector(A_prior),sqrt(as.vector(theta.prop)),log=TRUE))+dgamma(shrink.1.prop,0.01,0.01,log=TRUE)
    post1 <-   sum(dnorm(as.vector(A_draw),as.vector(A_prior),sqrt(as.vector(theta)),log=TRUE))+dgamma(shrink.1,.01,0.01,log=TRUE)
    if ((post1.prop-post1)>log(runif(1,0,1))){
      shrink.1 <- shrink.1.prop
      theta <- theta.prop
      accept <- accept+1
    }
    #Step for the second shrinkage parameter
    #Step for the first shrinkage parameter
    shrink.2.prop <- exp(rnorm(1,0,scale2))*shrink.2
    theta.prop <- get_V(shrink.1,shrink.2.prop,ind1=ind,sigma_sq1=sigma_sq,p1=p,set.cons=cons,M1=M,K1=K)
    
    post2.prop <- sum(dnorm(as.vector(A_draw),as.vector(A_prior),sqrt(as.vector(theta.prop)),log=TRUE))+dgamma(shrink.2.prop,0.01,0.01,log=TRUE)
    post2 <-   sum(dnorm(as.vector(A_draw),as.vector(A_prior),sqrt(as.vector(theta)),log=TRUE))+dgamma(shrink.2,0.01,0.01,log=TRUE)
    if ((post2.prop-post2)>log(runif(1,0,1))){
      shrink.2 <- shrink.2.prop
      theta <- theta.prop
      accept2 <- accept2+1
    }
    
    if (irep<(0.5*nburn)){
      if ((accept/irep)>0.3) scale1 <- 1.01*scale1
      if ((accept/irep)<0.15) scale1 <- 0.99*scale1
      if ((accept2/irep)>0.3) scale2 <- 1.01*scale2
      if ((accept2/irep)<0.15) scale2 <- 0.99*scale2
    }
    
  }
  }
    #Step IV: Sample covariance parameters
    #Get orthogonalized errors
    #Sample stochastic volatilities
    if (sv){
      pars_var <- matrix(0,3,M)
      for (jj in 1:M){
        svdraw <- svsample2(Em.str[,jj],startpara=para(svl[[jj]]),startlatent=H[,jj],priorphi = c(25, 1.5))
        svl[[jj]] <- svdraw
        h_ <- exp(svdraw$latent)
        pars_var[,jj] <- para(svl[[jj]])
        H[,jj] <- log(h_)
        SIGMA <- Upsilon_draw%*%diag(exp(H[T,]))%*%t(Upsilon_draw)
      }
    }else{
      for (jj in 1:M){
        S_1 <- a_i/2+T/2
        S_2 <- b_i/2+crossprod(Em.str[,jj])/2
        
        sig_eta <- 1/rgamma(1,S_1,S_2)
        H[,jj] <- log(sig_eta)
        SIGMA <- Upsilon_draw%*%diag(exp(H[1,]))%*%t(Upsilon_draw)
      }
    }
    # Create SIGMA
    eta_0 <- unlist(eta)
    
    if (shrink.type=="SSVS"){
    #Step VI: Sample the prior scaling factors for covariances from GIG
    M1 <- 0
    for (nn in 2:M){
      M1 <-M1+1 
      et_0 <- eta[[nn]]
      for (jj in  1:M1){
        u_ij1  <-  (1/kappa_0)*exp(-0.5*((et_0[[jj]])^2)/((kappa_0)^2))*q_ij
        u_ij2  <-  (1/kappa_1)*exp(-0.5*((et_0[[jj]])^2)/((kappa_1)^2))*(1-q_ij)
        ost  <-  u_ij1/(u_ij1 + u_ij2)
        if (is.na(ost)) ost <- 1
        sample_omega <-  bernoulli(ost)
        if (is.na(sample_omega)) sample_omega <- 1
        if (sample_omega==1){
          b_draw[nn,jj] <- kappa_1^2
        }else{
          b_draw[nn,jj] <- kappa_0^2
        }
      }
    }
    }
    #Step VIII: Store draws after burn-in/ Compute forecasts/ Impulse responses etc.
    if (irep>nburn){
      if (store.draws){
        A_store[irep-nburn,,] <- A_draw
        SIGMA_store[irep-nburn,,] <- SIGMA
        theta_store[irep-nburn,,] <- matrix(theta,K,M)
        omega_store[irep-nburn,,] <- b_draw
        Upsilon_store[irep-nburn,,] <- solve(Upsilon_draw)
        if (shrink.type=="SSVS") gamma_store[irep-nburn,,] <- matrix(gammas,K,M)
        if (shrink.type=="Minnesota")  shrink_store[irep-nburn,] <- c(shrink.1,shrink.2)
        if (sv){
          pars_store[irep-nburn,,] <- pars_var
          hv_store[irep-nburn,,] <- H
        } 
      }
      
      Lik <- 0
      for (t in 1:T) Lik <- Lik+dmvnorm(Y[t,],X[t,]%*%A_draw,Upsilon_draw%*%diag(exp(H[t,]))%*%t(Upsilon_draw),log=TRUE)
      
      if (method=="LIK"){
        Lik_store[irep-nburn,] <-  Lik#
      }else{
        Lik_store[irep-nburn,] <-  part.filt(n.part=1000,H1=H,Upsilon_draw1=Upsilon_draw,A_draw1=A_draw,y=Y,X1=X,pars_var1=pars_var)
      }
      
      #a_store[irep-nburn,] <- c(a_tau,b_tau)
      if (fcst==1){
        #Compute draw from the predictive density
        zt <- Xraw[nrow(Xraw)-h,]
        z1 <- zt
        Mean00 <- zt
        Sigma00 <- matrix(0,K,K)
        Sigma00[1:M,1:M] <- SIGMA
        Cm <- diag(M)
        varndxv <- c(M,cons,p)
        Htt <- (H[T,])
        SIGMAt <- SIGMA
        companion <- get_companion(A_draw,varndxv)
        Jm <- companion$Jm;Mm <- companion$MM
        #if (max(abs(Re(eigen(Mm)$values)))>1.05) next
        for (ih in 1:fhorz){
          z1 <- Mm%*%z1
          cholSig <- try(t(chol(Sigma00[1:M,1:M])),silent=TRUE)
          if (is(cholSig,"try-error")){
            yf <- rmvnorm(1,z1[1:M],Sigma00[1:M,1:M])
          }else{
            yf <- z1[1:M]+cholSig%*%rnorm(M,0,1)
          }
          #first and second moments
          Mean00 <- Mm%*%Mean00
          Sigma00 <- Mm%*%Sigma00%*%t(Mm)+Jm%*%SIGMAt%*%t(Jm)
          meanyt <- Cm%*%Mean00[1:M]
          varyt <- Cm%*%Sigma00[1:M,1:M]%*%t(Cm)
          pred_store[irep-nburn,,ih] <- yf
          if (sv){
            Htt <- pars_var[1,]+pars_var[2,]*(Htt-pars_var[1,])+rnorm(M,0,sqrt(pars_var[3,]))
            SIGMAt <- Upsilon_draw%*%diag(exp(Htt))%*%t(Upsilon_draw)
          }
        }
      }
      if (imps==1){
        varndxv <- c(M,cons,p)
        companion <- get_companion(A_draw,varndxv)
        
        if (max(abs(Re(eigen(companion$MM)$values)))>1.05) next
        PHI_array <- array(0,c(M,M,p))
        for (ss in 1:p){
          PHI_array[,,ss] <- t(A_draw[((ss-1)*M+1):(ss*M),])
        }
        irf_0 <- impulsdtrf(PHI_array,t(chol(SIGMA)),imphor)
        IRF_store[irep-nburn,,,] <- irf_0
      }
      
    }
    # setTxtProgressBar(pb, irep)
  }
  end <- Sys.time()
  print(end-start)   
  
  if (store.draws){
    A_post <- apply(A_store,c(2,3),median)
    A_mean <- apply(A_store,c(2,3),mean)
    
    SIGMA_post <- apply(SIGMA_store,c(2,3),median)
    theta_post <- apply(theta_store,c(2,3),median)
    omega_post <- apply(omega_store,c(2,3),median)
    Upsilon_post <- apply(Upsilon_store,c(2,3),median)
    Upsilon_draw <- apply(Upsilon_store,c(2,3),mean)
    DIC_OA <- NULL
    if (sv){
      H_post <- apply(exp(hv_store),c(2,3),mean)
      pars_post <- apply(pars_store,c(2,3),mean)
      Lik.post <- 0
      for (t in 1:T) Lik.post <- Lik+dmvnorm(Y[t,],X[t,]%*%A_mean,solve(Upsilon_draw)%*%diag((H_post[t,]))%*%t(solve(Upsilon_draw)),log=TRUE)
      
      if (method=="PF"){
        Lik.post <- part.filt(n.part=3000,H1=H_post,Upsilon_draw1=solve(Upsilon_draw),A_draw1=A_mean,y=Y,X1=X,pars_var1=pars_post)#part.filt(n.part=3000,H1=H_post,Upsilon_draw1=Upsilon_mean,A_draw1=A_mean,pars_var1 = pars_post,y=Y,X1=X)
      }
      
      Dbar <- mean(-4*Lik_store)
      pD <-2*Lik.post
      
      DIC_OA <- c(Dbar+pD,Dbar,pD)
      
    } 
    
    #Lik.post <- part.filt(n.part=3000,H1=H_post,Upsilon_draw1=solve(Upsilon_draw),A_draw1=A_mean,y=Y,X1=X,pars_var1=pars_post)#part.filt(n.part=3000,H1=H_post,Upsilon_draw1=Upsilon_mean,A_draw1=A_mean,pars_var1 = pars_post,y=Y,X1=X)
    
    

    
    return.list <- list(A=A_store,A_mean=A_post,SIGMA=SIGMA_store,SIGMA_mean=SIGMA_post,a=a_store,Upsilon_mean=Upsilon_post,Upsilon=Upsilon_store,theta=theta_store,theta_mean=theta_post,omega=omega_store,omega_mean=omega_post,predictions=pred_store,A_OLS=A_OLS,SIGMA_OLS=SIGMA_OLS,DIC=DIC_OA,IRF=IRF_store)
  }else{
    return.list <- list(predictions=pred_store,a=a_store)
  }
  
 return(return.list)
 }



